/*************************************************************************
 *
*    Used with ICCARM and AARM.
 *
 *    (c) Copyright IAR Systems 2007
 *
 *    File name   : sys.c
 *    Description : System module
 *
 *    History :
 *    1. Date        : 13, February 2008
 *       Author      : Stanimir Bonev
 *       Description : Create
 *
 *
 *    $Revision: 30870 $
 **************************************************************************/
#include "types.h"
#include "sys.h"

#ifndef __FIQ_USED
/*************************************************************************
 * Function Name: fiq_handler
 * Parameters: none
 *
 * Return: none
 *
 * Description: FIQ handler
 *
 *************************************************************************/
__fiq __arm void FIQ_Handler (void)
{
	while(1);
}
#endif // __FIQ_USED

#ifndef __IRQ_USED
/*************************************************************************
 * Function Name: irq_handler
 * Parameters: none
 *
 * Return: none
 *
 * Description: IRQ handler
 *
 *************************************************************************/
void IRQ_Handler (void)
{
	FNCT_PTR interruptFunction;

	// Get interrupt vector.
	interruptFunction = (FNCT_PTR)VICADDRESS;

	if(interruptFunction != NULL)
	{
		// Call vectored interrupt function.
		interruptFunction();

		// Acknowledge VIC
		VICADDRESS = 1;
	}
	else
	{
		// Clear interrupt in VIC.
		VICADDRESS = 0;
	}
}
#endif // __IRQ_USED

/*************************************************************************
 * Function Name: VIC_Init
 * Parameters: void
 * Return: void
 *
 * Description: Initialize VIC
 *
 *************************************************************************/
void VIC_Init(void)
{
	volatile unsigned long * pVecAdd, *pVecCntl;
	int i;
	// Assign all interrupt channels to IRQ
	VICINTSELECT  =  0;
	// Disable all interrupts
	VICINTENCLEAR = 0xFFFFFFFF;
	// Clear all software interrupts
	VICSOFTINTCLEAR = 0xFFFFFFFF;
	// VIC registers can be accessed in User or privileged mode
	VICPROTECTION = 0;
	// Clear interrupt
	VICADDRESS = 0;

	// Clear address of the Interrupt Service routine (ISR) for vectored IRQs
	// and disable all vectored IRQ slots

	pVecAdd = (volatile unsigned long *)&VICVECTADDR0;
	pVecCntl = (volatile unsigned long *)&VICVECTPRIORITY0;

	for(i = 0; i < 32; ++i)
	{
		*pVecAdd++ = 0;
		*pVecCntl++ = 0;
	}
}

/*************************************************************************
 * Function Name: VIC_SetVectoredIRQ
 * Parameters:  void(*pIRQSub)()
 *              unsigned int VicIrqSlot
 *              unsigned int VicIntSouce
 *
 * Return: void
 *
 * Description:  Init vectored interrupts
 *
 *************************************************************************/
void VIC_SetVectoredIRQ(void(*pIRQSub)(void), unsigned int Priority, unsigned int VicIntSource)
{
	unsigned long volatile *pReg;
	// load base address of vectored address registers
	pReg = (unsigned long volatile *)&VICVECTADDR0;

	// Set the vector address to the handler's address
	*(pReg + VicIntSource) = (unsigned long)pIRQSub;

	// load base address of ctrl registers
	pReg = (unsigned long volatile *)&VICVECTPRIORITY0;

	// Set source channel and enable the slot
	*(pReg + VicIntSource) = Priority;

	// Clear FIQ select bit
	VICINTSELECT &= ~( 1 << VicIntSource);
}

/*************************************************************************
 * Function Name: InitClock
 * Parameters: void
 * Return: void
 *
 * Description: Initialize PLL and clocks' dividers. Hclk - 288MHz,
 * Cclk- 72MHz, Usbclk - 48MHz
 *
 *************************************************************************/
void InitClock(void)
{
#if 0
	// 1. Init OSC
	SCS_bit.OSCRANGE = 0;
	SCS_bit.OSCEN = 1;
	// 2.  Wait for OSC ready
	while(!SCS_bit.OSCSTAT);

	// 3. Disconnect PLL
	PLLCON_bit.PLLC = 0;
	PLLFEED = 0xAA;
	PLLFEED = 0x55;

	// 4. Disable PLL
	PLLCON_bit.PLLE = 0;
	PLLFEED = 0xAA;
	PLLFEED = 0x55;

	// 5. Select source clock for PLL
	CLKSRCSEL_bit.CLKSRC = 1; // Selects the main oscillator as a PLL clock source.

	// 6. Set PLL settings 288 MHz
	PLLCFG_bit.MSEL = 12-1;
	PLLCFG_bit.NSEL = 1-1;
	PLLFEED = 0xAA;
	PLLFEED = 0x55;

	// 7. Enable PLL
	PLLCON_bit.PLLE = 1;
	PLLFEED = 0xAA;
	PLLFEED = 0x55;

	// 9. Set clk divider settings
	CCLKCFG   = 4-1;            // 1/4 Fpll - 72 MHz
	USBCLKCFG = 6-1;            // 1/6 Fpll - 48 MHz
	PCLKSEL0 = PCLKSEL1 = 0;    // other peripherals - 18MHz (Fcclk/4)

	// 8. Wait for the PLL to achieve lock
	while(!PLLSTAT_bit.PLOCK);

	// 10. Connect the PLL
	PLLCON_bit.PLLC = 1;
	PLLFEED = 0xAA;
	PLLFEED = 0x55;
#else
    /* [1] Check if PLL connected, disconnect if yes. */
    if (PLLSTAT & (1<<25)/* PLLSTAT_PLLC*/)
    {
        PLLCON = (1<<0) /*PLLCON_PLLE*/ ;
        /* Enable PLL, disconnected ( PLLC = 0)*/
        PLLFEED = 0xAA;
        PLLFEED = 0x55;
    }

    /* [2] Disable the PLL once it has been disconnected. */
    PLLCON  = 0;
    PLLFEED = 0xAA;
    PLLFEED = 0x55;

    /* [3] Change the CPU Clock Divider setting
     * to speed up operation without the PLL, if desired.
     * We're going to divide by 1 for maximum non-PLL clock speed.
     * NOTE: CCLKCFG adds one internally so we must subtract one */
    CCLKCFG = ( (1) /* CCLK_DIV_1 */ - 1);

    /* [4] Enable the main oscillator, select clock source  */
    if( (12000000) /*FOSC_MAIN*/ > 20000000 )
    {
        SCS |= (1<<4) /*SCS_OSCRANGE*/;
    }
    else
    {
        SCS &= ~(1<<4) /*SCS_OSCRANGE*/;
    }
    SCS |= (1<<5) /*SCS_OSCEN*/;

    /* Wait until main OSC is usable */
    do {} while ((SCS & (1<<6) /*SCS_OSCSTAT*/) == 0);

    /* Select the main osc as the PLL clk source. */
    CLKSRCSEL = (0x01) /*CLKSRC_MAIN_OSC*/;

    /*
     * NOTES:
     *
     * Set multiplier and divider values.
     *
     *  PLLCFG = ((N - 1) << 16) + (M - 1)
     *
     *  F_cco = (2 * M *F_in)/N
     *      M = 12, N = 1, F_in = 12.000 MHz -> F_cco = 288.000 MHz
     *
     */
    /* [5] Write to the PLLCFG and make it effective with feed sequence */
    uint32_t pllcfg = (((1 - 1) << 16) + ((12 - 1) << 0));
    PLLCFG = pllcfg;
    PLLFEED = 0xAA;
    PLLFEED = 0x55;

    /* [6] Enable the PLL and make it effective with feed sequence */
    PLLCON  = (1<<0) /*PLLCON_PLLE*/;
    PLLFEED = 0xAA;
    PLLFEED = 0x55;

    /* [7] Change the CPU Clock Divider setting for the operation with the PLL.
     *     It's critical to do this before connecting the PLL.
     * NOTE: CCLKCFG adds one internally so we must subtract one
     * Divide F_cco down to get the CCLK output. */
    CCLKCFG = (4 - 1);
	USBCLKCFG = 6-1;            // 1/6 Fpll - 48 MHz

    /* [8] Wait for the PLL to lock to set frequency
     * TODO extra credit: Add timeout for slow FREF */
    do {} while(((PLLSTAT) & (1<<26) /*PLLSTAT_PLOCK*/) == 0);
    do {} while(((PLLSTAT) & 0x00FF7FFF) != pllcfg);

    /* [9] Enable and connect the PLL as the clock source */
    PLLCON = ((1<<0) /*PLLCON_PLLE*/ | (1<<1)/*PLLCON_PLLC*/);
    PLLFEED = 0xAA;
    PLLFEED = 0x55;

    /* Check connect bit status and wait for connection. */
    do {} while(((PLLSTAT) & (1<<25)/*PLLSTAT_PLLC*/) == 0);

#endif
	// stop all Peripherals' clocks
	PCONP = 0;

	// Configure AHB1 arbitration
	AHBCFG1_bit.SHDL = 0;     // priority arbitration
	AHBCFG1_bit.BB   = 0;     // Never break defined length bursts.
	AHBCFG1_bit.QT   = 0;     // A quantum is an AHB clock.
	AHBCFG1_bit.QS   = 4;     // Preemptive, re-arbitrate after 16 AHB quanta.
	AHBCFG1_bit.DM   = 1;     // Master 1 (CPU) is the default master.
	AHBCFG1_bit.EP1  = 1;     // External priority for master 1 (CPU).
	AHBCFG1_bit.EP2  = 2;     // External priority for master 2 (GPDMA).
	AHBCFG1_bit.EP3  = 3;     // External priority for master 3 (AHB1).
	AHBCFG1_bit.EP4  = 5;     // External priority for master 4 (USB).
	AHBCFG1_bit.EP5  = 4;     // External priority for master 5 (LCD).
}

/*************************************************************************
 * Function Name: SYS_GetFsclk
 * Parameters: none
 * Return: Int32U
 *
 * Description: return Sclk [Hz]
 *
 *************************************************************************/
Int32U SYS_GetFsclk(void)
{
	Int32U Mul = 1, Div = 1, Osc, Fsclk;
	if(PLLSTAT_bit.PLLC && PLLSTAT_bit.PLLE)
	{
		// when PLL is connected
		Mul = PLLSTAT_bit.MSEL + 1;
		Div = PLLSTAT_bit.NSEL + 1;
	}

	// Find clk source
	switch(CLKSRCSEL_bit.CLKSRC)
	{
		case 0:
			Osc = I_RC_OSC_FREQ;
			break;
		case 1:
			Osc = MAIN_OSC_FREQ;
			break;
		case 2:
			Osc = RTC_OSC_FREQ;
			break;
		default:
			Osc = 0;
	}
	// Calculate system frequency
	Fsclk = Osc*Mul*2;
	Fsclk /= Div*(CCLKCFG+1);
	return(Fsclk);
}

/*************************************************************************
 * Function Name: SYS_GetFpclk
 * Parameters: Int32U Periphery
 * Return: Int32U
 *
 * Description: return Pclk [Hz]
 *
 *************************************************************************/
Int32U SYS_GetFpclk(Int32U Periphery)
{
	Int32U Fpclk;
	pInt32U pReg = (pInt32U)((Periphery < 32) ? &PCLKSEL0 : &PCLKSEL1);

	Periphery  &= 0x1F;   // %32
	Fpclk = SYS_GetFsclk();
	// find peripheral appropriate periphery divider
	switch((*pReg >> Periphery) & 3)
	{
		case 0:
			Fpclk /= 4;
			break;
		case 1:
			break;
		case 2:
			Fpclk /= 2;
			break;
		default:
			Fpclk /= 8;
	}
	return(Fpclk);
}

/*************************************************************************
 * Function Name: GpioInit
 * Parameters: void
 * Return: void
 *
 * Description: Reset all GPIO pins to default: primary function
 *
 *************************************************************************/
void GpioInit(void)
{
	// Set to inputs
	IO0DIR  = 0;
	IO1DIR  = 0;
	FIO0DIR = 0;
	FIO1DIR = 0;
	FIO2DIR = 0;
	FIO3DIR = 0;
	FIO4DIR = 0;

	// Enable Fast GPIO0,1
	SCS_bit.GPIOM = 1;

	// clear mask registers
	FIO0MASK = 0;
	FIO1MASK = 0;
	FIO2MASK = 0;
	FIO3MASK = 0;
	FIO4MASK = 0;

#ifndef SDRAM_DEBUG
	// Reset all GPIO pins to default primary function
	PINSEL0 = 0;
	PINSEL1 = 0;
	PINSEL2 = 0;
	PINSEL3 = 0;
	PINSEL4 = 0;
	PINSEL5 = 0;
	PINSEL6 = 0;
	PINSEL7 = 0;
	PINSEL8 = 0;
	PINSEL9 = 0;
	PINSEL10= 0;
#endif
}

